from importlib import import_module

from ordotools.tools.feast import Feast

from ordotools.tools.helpers import LiturgicalYearMarks
from ordotools.tools.helpers import days
from ordotools.tools.helpers import ladys_office
from ordotools.tools.helpers import leap_year

from ordotools.tools.liturgical_dates import integer_to_roman

from ordotools.tools.temporal import Temporal

from ordotools.sanctoral.diocese.roman import Sanctoral

import logging
from logging import debug
from logging import info

logging.basicConfig(level=logging.DEBUG)


class LiturgicalCalendar:

    def __init__(self, year, diocese, country=""):
        self.year = year
        self.diocese = diocese
        self.transfers = None
        self.temporal = Temporal(self.year).return_temporal()
        self.FIRST_ADVENT = LiturgicalYearMarks(self.year).first_advent
        self.LAST_ADVENT = LiturgicalYearMarks(self.year).last_advent
        self.LENT_BEGINS = LiturgicalYearMarks(self.year).lent_begins
        self.LENT_ENDS = LiturgicalYearMarks(self.year).lent_ends

    def explode_octaves(self, feast: Feast) -> dict:
        """
        Generates "explodes" the octave for a feast.
        """
        octave = {}
        for x in range(7):
            intro = f"De {integer_to_roman(x+2)} die"
            if x != 6:
                feast.name = f"{intro} infra {feast.infra_octave_name}"
            else:
                feast.name = feast.infra_octave_name
            feast.rank_v = "feria"
            feast.rank_n = 18 if x < 6 else 13  # common octave
            octave |= {feast.date+days(x+1): feast.updated_properties}
        return octave

    def find_octave(self, year: dict) -> dict:
        """
        Finds the octaves of the sanctoral cycle and adds
        them to the year.
        """
        y = year.copy()
        temporals = []
        for feast in self.temporal.values():
            temporals.append(feast["feast"])
        for date, feast in y.items():
            candidate = Feast(feast_date=date, properties=feast)
            if candidate.name in temporals:
                continue
            elif candidate.octave is True:
                octave = self.explode_octaves(candidate)
                y |= self.add_feasts(master=y, addition=octave)
        return y

    def commemoration(self, feast: Feast, com: Feast) -> Feast:
        feast.com.insert(0, com.feast_properties)
        return feast

    def rank_by_nobility(self, feast_1: Feast, feast_2: Feast) -> dict:
        for x in range(6):
            if feast_1.nobility[x] < feast_2.nobility[x]:
                return {'higher': feast_1, 'lower': feast_2}
            elif feast_1.nobility[x] > feast_2.nobility[x]:
                return {'higher': feast_2, 'lower': feast_1}
            else:
                pass
        return {'higher': feast_1, 'lower': feast_2}

    def rank(self, dynamic: Feast, static: Feast) -> Feast:
        """
        Ranks feasts that occur on the same date. Send feasts that can be
        transferred to the transfer dictionary. "Dynamic" is from the
        sanctoral cycle, and "static" is from the temporal cycle.
        """
        info(f'ranking "{dynamic.feast}" against "{static.feast}"')
        # if the feasts are the same rank:
        if dynamic.rank_n == static.rank_n:
            # it is possible that the lower can be the
            # commemoration of the higher...
            return self.rank_by_nobility(
                    dynamic,
                    static
                )['higher']
        else:
            candidates = {
                dynamic.rank_n: dynamic,
                static.rank_n: static,
            }
            higher = candidates[sorted(candidates)[0]]
            lower = candidates[sorted(candidates)[1]]
            if lower.rank_n == 22:
                pass
            if lower.rank_n == 19:  # lent
                return self.commemoration(
                        feast=higher,
                        com=lower
                    )

            # feasts that do not take a commemoration
            if higher.rank_n <= 4:
                # lower feast is transferred
                if lower.rank_n == 3:  # holy week ferias, etc.
                    self.transfers == higher
                    return lower
                if lower.rank_n <= 10:
                    if lower.fasting is True:
                        higher.fasting = True
                    if lower != self.transfers:
                        self.transfers = lower
                    return higher
                else:
                    # lower feast is ignored
                    if lower.fasting is True:
                        higher.fasting = True
                    return higher

            # impeded DM, D and SD
            elif 14 <= lower.rank_n <= 16:
                return self.commemoration(
                        feast=higher,
                        com=lower
                    )
            else:
                return self.commemoration(
                        feast=higher,
                        com=lower
                    )
        # return ranked_feasts

    def transfer_feast(self, feast: Feast) -> None:
        """
        Checks for feasts in the transfer dictionary.
        """
        info(f"{self.transfers.feast} is being transferred.")
        return self.rank(dynamic=self.transfers, static=feast)

    def add_feasts(self, master: dict, addition: dict) -> dict:
        """
        Adds additional feasts to the master feast dictionary;
        Sends the feasts that confilict to the ranking algorithm.
        """
        calendar = master.copy()
        for date, data in calendar.items():
            if date in addition:
                feast = self.rank(
                    dynamic=Feast(date, addition[date]),
                    static=Feast(date, data)
                )
            else:
                feast = Feast(date, data)
            if self.transfers is not None:
                result = self.transfer_feast(
                    Feast(date, feast.updated_properties)
                )
                if result.feast == feast.feast:
                    pass
                else:
                    self.transfers = None
                    feast = result
            try: # this might not be needed anymore
                # calendar |= {date: feast.updated_properties}
                calendar |= {date: feast}
            except TypeError:
                pass
        return calendar

    def our_ladys_saturday(self, calendar: dict) -> None:
        """
        Adds Office of the BVM to the compiled calendar.
        """
        # TODO: add Mass number according to season
        year = calendar.copy()
        office = ladys_office
        for feast in year.keys():
            if feast.strftime("%w") == str(6):
                if self.LENT_BEGINS.date() <= feast.date() <= self.LENT_ENDS.date():
                    continue
                elif self.FIRST_ADVENT.date() <= feast.date() <= self.LAST_ADVENT.date():
                    continue
                else:
                    if year[feast].rank_n > 16:  # not a double
                        year[feast] = Feast(feast, office)  # TODO: add commemorations
                    else:
                        continue
        return year

    def build(self) -> None:
        saints = Sanctoral(self.year)
        if self.diocese == 'roman':
            sanctoral = saints.data if leap_year(self.year) is False \
                else saints.leapyear()
        else:
            diocese = import_module(
                f"sanctoral.diocese.{self.diocese}",
            )
            sanctoral = diocese.Diocese(self.year).calendar()
        full_calendar = self.add_feasts(self.temporal, sanctoral).copy()
        full_calendar |= self.our_ladys_saturday(full_calendar)
        full_calendar |= self.find_octave(year=full_calendar)
        return full_calendar



from datetime import datetime


class Feast:
    def __init__(self, feast_date: datetime, properties: dict):
        self.date = feast_date
        self.feast_properties = properties
        self.name = properties['feast']
        if 'infra_octave_name' in properties.keys():
            self.infra_octave_name = properties['infra_octave_name']
        else:
            self.infra_octave_name = self.name
        self.rank_v = properties['rank'][-1]  # verbose rank
        self.rank_n = properties['rank'][0]   # numeric rank
        self.octave = True if 'Oct' in self.rank_v else False
        self.color = properties['color']
        self.mass = {}
        if 'int' in properties['mass'].keys():
            self.mass = {'Ad Missam': properties['mass']}
        else:
            for x, y in properties['mass'].items():
                self.mass.update({x: y})
        self.nobility = properties['nobility'] \
            if 'nobility' in properties.keys() else (0, 0, 0, 0, 0, 0,)
        self.office_type = properties['office_type']
        self.com = properties['com'] \
            if 'com' in properties.keys() else []
        self.matins = properties['matins'] \
            if 'matins' in properties.keys() else {}
        self.lauds = properties['lauds'] \
            if 'lauds' in properties.keys() else {}
        self.prime = properties['prime'] \
            if 'prime' in properties.keys() else {}
        self.little_hours = properties['little_hours'] \
            if 'little_hours' in properties.keys() else {}
        self.vespers = properties['vespers'] \
            if 'vespers' in properties.keys() else {}
        self.compline = properties['compline'] \
            if 'compline' in properties.keys() else {}
        self.fasting = properties['fasting']

    @ property
    def feast_date_display(self) -> str:
        """ Displays the feast date for the ordo """
        return self.date.strftime('%d')

    @ property
    def readme_date(self) -> str:
        """ Displays the feast date for the ordo """
        return self.date.strftime('%b %d')

    # def date(self):
    #     # ? is this used?
    #     return datetime.strptime('%m%d', self.feast_date)

    @ property
    # TODO: expand Preces method to provide for
    #    1. Preces Feriales
    #    2. Compline
    #    3. Little Hours
    def preces(self) -> str:
        if self.rank_n <= 16:
            return ''
        else:
            return 'Preces'

    @ property
    def feast(self) -> str:
        return self.feast_properties['feast']

    @ property
    def updated_properties(self) -> dict:
        """ Updates all values of the feast's dictionary """
        if self.com and "fasting" in self.com[0]:
            if self.com[0]["fasting"]:
                fasting = self.com[0]["fasting"]
            else:
                fasting = self.fasting
        elif int(self.date.strftime("%m")) == 5:
            fasting = True
        else:
            fasting = self.fasting
        dic = {
            "feast": self.name,
            "rank": [self.rank_n, self.rank_v],
            "color": self.color,
            "mass": self.mass,
            "matins": {},
            "lauds": {},
            "prime": {},
            "little_hours": {},
            "vespers": self.vespers,
            "compline": {},
            "nobility": self.nobility,
            "office_type": self.office_type,
            "com": self.com,
            "fasting": fasting,
        }
        return dic

    @ property
    def office_type2latex(self) -> str:
        """ returns formatted office type based on office_type key """
        # todo write all office types in english
        if self.office_type is False:
            off_type = 'Ord, '
        elif self.office_type == 'feria':
            off_type = 'Fer, '
        elif self.office_type == 'festiva':
            off_type = 'Festiv, '
        elif self.office_type == 'dominica':
            off_type = 'Dom, '
        elif self.office_type == 'ut in pr loco':
            off_type = 'ut in pr loco, '
        else:
            return 'ERROR!'
        return 'Off ' + off_type

    @ property
    def com_in_title(self) -> str:
        """ returns formatted commemorations in title """
        results = ''
        for x in self.com:
            if type(x) == dict:
                if 'feast' in x.keys():
                    results += r'\textit{['+x['feast'] + r']} \\ '
                else:
                    results += ''
            else:
                results += ''
        return results

    def introit(self) -> list:
        """ Finds the Introit text for the feast """
        # TODO: add a methed to determine if it is a three Mass Introit or not.
        introit_list = []
        if len(self.mass.items()) == 1:
            if 'Ad Missam' in self.mass.keys():
                if type(self.mass['Ad Missam']['int']) == str:
                    introit_list.append(self.mass['Ad Missam']['int'])
                else:
                    # TODO: determine whether it is still pascahltime or not
                    # ! just for testing purposes
                    introit_list.append(self.mass['Ad Missam']['int'][0])
            else:
                pass
        else:
            for y in self.mass.keys():
                introit_list.append(self.mass[y]['int'])
        return introit_list

    @ property
    def translate_color(self) -> str:
        """ translate the color from english to latin """
        translations = {
            "white": "albus",
            "green": "viridis",
            "red": "ruber",
            "purple": "violaceus",
            "black": "niger",
            "pink": "rosa",
            "color": "void",
        }
        latin_color = translations[self.color]
        return latin_color

    @ property
    def translate_weekday(self) -> str:
        """ translate the weekday from english to latin """
        translations = {
            'Sunday': 'Dom',
            'Monday': 'Fer II',
            'Tuesday': 'Fer III',
            'Wednesday': 'Fer IV',
            'Thursday': 'Fer V',
            'Friday': 'Fer VI',
            'Saturday': 'Sabb',
        }
        # latin_weekday = translations[datetime.strptime(self.feast_date.strip(
        #     'tranlsated ._')+'/'+str(YEAR), '%m/%d/%Y').strftime('%A')]
        latin_weekday = translations[self.date.strftime('%A')]
        return latin_weekday

    def mass_commemorations(self) -> str:
        """ returns formatted and ordered mass commemorations in title """
        results = ''
        i = 2
        for x in self.com:
            if type(x) == dict:
                if 'feast' in x.keys():
                    results += str(i) + r' or ' + x['feast'] + r', '
                    i += 1
                # TODO: use 'feast' data as well
                elif 'oration' in x.keys():
                    if x['oration'] == 'ad libitum':
                        results += str(i) + r' or ' + \
                            x['oration'] + r', '
                    else:
                        results += str(i) + \
                            r' or \textit{' + x['oration'] + r',} '
                    i += 1
                else:
                    pass
            else:
                pass
        return results

    def display_mass_as_latex(self) -> str:
        """ return the Mass as LaTeX code """
        # todo add orations
        latexed_mass = ''
        pprlast = ''
        status = []
        for y in self.mass.values():
            gloria = y['glo']
            creed = y['cre']
            if gloria is True and creed is True:
                status.append('G, C, ')
            elif gloria is True and creed is False:
                status.append('G, ')
            elif gloria is False and creed is True:
                status.append('C, ')
            else:
                status.append('')
        i = 0
        # FIX: refine this to work only for its proper Mass:
        if 'proper_last_gospel' in y.keys():
            pprlast = 'ult Evgl ' + y['proper_last_gospel'] + ', '
        else:
            pprlast = ''
        for x, y in self.mass.items():
            # TODO: use the second in the string if it is Paschaltime.
            latexed_mass += '\\textbf{'+x+'}: \\textit{' + \
                self.introit()[i] + ',} ' + \
                status[i]+'Pre '+y['pre'] + ', ' + \
                self.mass_commemorations() + ' ' + pprlast + ' '
            i += 1
        return latexed_mass

    @ property
    def display_matins_as_latex(self) -> str:
        """ return Matins as LaTeX code """
        if len(self.matins) == 0:
            return ""
        else:
            return r"\textbf{Ad Mat}: "

    @ property
    def display_lauds_as_latex(self) -> str:
        """ return Laudes as LaTeX code """
        if len(self.lauds) == 0:
            return ''
        else:
            latexed_lauds = r'\textbf{Ad Lau}: '
            for k in self.lauds.keys():
                if k == 'psalms':
                    # * make this a bit easier to work with
                    if self.lauds[k] == 'sunday':
                        latexed_lauds += 'Pss de Dom, '
                    else:
                        pass
            return latexed_lauds

    @ property
    def display_little_hours_as_latex(self) -> str:
        """ return the little hours as LaTeX code """
        if len(self.little_hours) == 0:
            return ''
        else:
            latexed_little_hours = r'\textbf{Ad Horas}: '
            for k in self.little_hours.keys():
                if k == 'preces_feriales':
                    if self.prime[k] is True:
                        latexed_little_hours += 'Preces feriales, '
                    else:
                        pass
                elif k == 'psalms':
                    if self.little_hours[k] == 'sunday':
                        latexed_little_hours += 'Pss de Dom, '
                else:
                    pass
            return latexed_little_hours

    @ property
    def display_prime_as_latex(self) -> str:
        """ return Prime as LaTeX code """
        if len(self.prime) == 0:
            return ''
        else:
            latexed_prime = r'\textbf{Ad Primam}: '
            for k in self.prime.keys():
                if k == 'four_psalms':
                    if self.prime[k] is True:
                        latexed_prime += r'4 Pss, '
                    else:
                        pass
                elif k == 'cap':
                    latexed_prime += r'\textit{' + f'{self.prime[k]},' + '} '
                elif k == 'v_r':
                    latexed_prime += r'\Vbar\ in \Rbar\ brev: ' + \
                        r'\textit{' + f'{self.prime[k]},' + '} '
                elif k == 'preces_feriales':
                    if self.prime[k] is True:
                        latexed_prime += 'Preces feriales, '
                    else:
                        pass
                else:
                    pass
            return latexed_prime

    @ property
    def display_vespers_as_latex(self) -> str:
        """ return vespers as LaTeX code """
        if len(self.vespers) == 0:
            return ""
        else:
            return r"\textbf{In Vesp}: "

    @ property
    def display_compline_as_latex(self) -> str:
        """ return compline as LaTeX code """
        if len(self.compline) == 0:
            return ""
        else:
            latexed_compline = r"\textbf{Ad Compl}: "
            for k in self.compline.keys():
                if k == "sunday":
                    if self.compline[k] is True:
                        latexed_compline += "Pss de Dom. "
                    else:
                        pass
                else:
                    pass
            return latexed_compline



from datetime import datetime
from math import floor


def integer_to_roman(num=int) -> int:
    romans = {
        1: "I", 5: "V", 10: "X", 50: "L", 100: "C",
        500: "D", 1000: "M", 5000: "G", 10000: "H",
    }
    div, result = 1, ''
    while num >= div:
        div *= 10
    div /= 10
    while num:
        last_num = int(num / div)
        if last_num <= 3:
            result += (romans[div]*last_num)
        elif last_num == 4:
            result += (romans[div]+romans[div*5])
        elif 5 <= last_num <= 8:
            result += (romans[div*5]+(romans[div]*(last_num-5)))
        elif last_num == 9:
            result += (romans[div]+romans[div*10])
        num = floor(num % div)
        div /= 10
    return result


def dominical(year=int) -> str:
    letter, y, c = '', int(str(year)[2:]), int(str(year)[:2])
    index = (y+(y/4)+5*(c % 4)-1) % 7
    letters = ['G', 'F', 'E', 'D', 'C', 'B', 'A']
    first_letter, second_letter = '', ''
    try:
        datetime(year, 2, 29)
        first_letter = letters[int(index)-1]
        if int(index) < 0:
            first_letter = letters[6]
        second_letter = letters[int(index)]
    except ValueError:
        second_letter = letters[int(index)]
    letter = first_letter+second_letter
    return letter


def golden_number(year=int) -> int:
    # todo remove the other golden number function in utils
    return (year+1) % 19


def epact_adjust(year=int) -> int:
    l, s, c = 0, 0, int(str(year)[:2])
    if c % 2:
        l = -1
    if not (c-2) % 4:
        l = -1
    if not c % 3:
        if not (c-17) % 25:
            s = 0
        else:
            s = 1
    if not (c-18) % 25:
        s = 1
    return l + s


def epact_build(adjust=int) -> list:
    build_base, day, x = [], 1, 0
    while x != 19:
        day += 11*(x if x == 0 else 1)
        if day > 30:
            day -= 30
        build_base.append(day)
        x += 1
    build = []
    for y in build_base:
        y += adjust
        if y > 31:
            y -= 31
        elif y <= 0:
            y += 31
        build.append(y)
    return build


def epact(year=int) -> str:
    # ? should we nest the other two functions here?
    e, base, i = '', 1600, 0
    while year >= base:
        i += epact_adjust(year)
        year -= 100
    epact_list = epact_build(i)
    e = integer_to_roman(epact_list[golden_number(year)-1])
    if e == 'XXXI':
        e = '*'
    return e


def epact_chart(year=int) -> int:
    """ This function is guaranteed to be accurate """
    gn = golden_number(year)
    epct = 0
    if 0 <= year <= 1582:
        epct = 0
    elif 1583 <= year <= 1699:
        epct = 1
    elif 1700 <= year <= 1899:
        epct = 2
    elif 1900 <= year <= 2199:
        epct = 3
    elif 2200 <= year <= 2299:
        epct = 4
    elif 2300 <= year <= 2399:
        epct = 5
    elif 2400 <= year <= 2499:
        epct = 6
    elif 2500 <= year <= 2599:
        epct = 7
    elif 2600 <= year <= 2899:
        epct = 8
    elif 2900 <= year <= 3099:
        epct = 9
    else:
        epct = 0
    matrix = [
        [30, 11, 22, 3, 14, 25, 6, 17, 28, 9, 20, 1, 12, 23, 4, 15, 26, 7, 18, ],
        [1, 12, 23, 4, 15, 26, 7, 18, 29, 10, 21, 2, 13, 24, 5, 16, 27, 8, 19, ],
        [30, 11, 22, 3, 14, 25, 6, 17, 28, 9, 20, 1, 12, 23, 4, 15, 26, 7, 18, ],
        [29, 10, 21, 2, 13, 24, 5, 16, 27, 8, 19, 30, 11, 22, 3, 14, 25, 6, 17, ],
        [28, 9, 20, 1, 12, 23, 4, 15, 26, 7, 18, 29, 10, 21, 2, 13, 24, 5, 16, ],
        [27, 8, 19, 30, 11, 22, 3, 14, 25, 6, 17, 28, 9, 20, 1, 12, 23, 4, 15, ],
        [28, 9, 20, 1, 12, 23, 4, 15, 26, 7, 18, 29, 10, 21, 2, 13, 24, 5, 16, ],
        [27, 8, 19, 30, 11, 22, 3, 14, 25, 6, 17, 28, 9, 20, 1, 12, 23, 4, 15, ],
        [26, 7, 18, 29, 10, 21, 2, 13, 24, 5, 16, 27, 8, 19, 30, 11, 22, 3, 14, ],
        [25, 6, 17, 28, 9, 20, 1, 12, 23, 4, 15, 26, 7, 18, 29, 10, 21, 2, 13, ],
    ]
    return matrix[epct][gn-1]

